import unittest
from unittest.mock import Mock

from mysql_mimic import types
from mysql_mimic.connection import Connection


class TestMysqlServer(unittest.TestCase):
    def setUp(self) -> None:
        self.conn = Connection(Mock(), Mock(), Mock())

    def test_ok(self):
        self.conn.capabilities = self.conn.server_capabilities
        self.assertEqual(self.conn.ok(), b"\x00\x00\x00\x00\x00\x00\x00")

    def test_ok_eof(self):
        self.conn.capabilities = self.conn.server_capabilities
        self.assertEqual(self.conn.ok(eof=True), b"\xfe\x00\x00\x00\x00\x00\x00")

    def test_ok_non_41(self):
        self.assertEqual(self.conn.ok(), b"\x00\x00\x00")

    def test_ok_transactions(self):
        self.conn.capabilities = types.Capabilities.CLIENT_TRANSACTIONS
        self.assertEqual(self.conn.ok(), b"\x00\x00\x00\x00\x00")

    def test_eof(self):
        self.conn.capabilities = self.conn.server_capabilities
        self.assertEqual(self.conn.eof(), b"\xFE\x00\x00\x00\x00")

    def test_eof_non_41(self):
        self.assertEqual(self.conn.eof(), b"\xFE")

    def test_empty_error(self):
        self.conn.capabilities = self.conn.server_capabilities
        self.assertEqual(self.conn.error(msg=""), b"\xFFQ\x04#HY000")

    def test_empty_error_non_41(self):
        self.assertEqual(self.conn.error(msg=""), b"\xFFQ\x04")

    def test_simple_error(self):
        self.conn.capabilities = self.conn.server_capabilities
        self.assertEqual(self.conn.error(msg="kelsin"), b"\xFFQ\x04#HY000kelsin")

    def test_column_definition_41(self):
        self.assertEqual(
            self.conn.column_definition_41(),
            b"\x03def\x00\x00\x00\x00\x00\x0c\xff\x00\x00\x01\x00\x00\x0f\x00\x00\x00\x00\x00",
        )



class Session:
    """
    Abstract client session.

    This should be implemented by applications.
    """

    async def query(self, sql):  # pylint: disable=unused-argument
        """
        Process a SQL query.

        Args:
            sql (str): SQL statement from client
        Returns:
            One of:
            - tuple(rows, column_names), where "rows" is a sequence of sequences
              and "column_names" is a sequence of strings with the same length
              as every row in "rows"
            - tuple(rows, result_columns), where "rows" is the same
              as above, and "result_columns" is a sequence of mysql_mimic.ResultColumn
              instances.
            - mysql_mimic.ResultSet instance
        """
        return [], []

    async def close(self):
        """
        Close the session.

        This is called when the client closes the connection
        """

    async def init(self, connection):
        """
        Initialize the session.

        Args:
            connection (Connection): connection of the session
        """

    async def set(self, **kwargs):
        """
        Set session variables.

        Args:
            **kwargs: mapping of variable names to values
        """

    async def show_columns(self, database, table):  # pylint: disable=unused-argument
        """
        Show column metadata.

        Args:
            database (str): database name
            table (str): table name
        Returns:
            list[mysql_mimic.results.Column|dict]: columns
        """
        return []



import io
import logging
import re
from dataclasses import dataclass
from typing import Optional, Dict, Iterable

from mysql_mimic.constants import DEFAULT_SERVER_CAPABILITIES
from mysql_mimic.errors import ErrorCode, MysqlError, get_sqlstate
from mysql_mimic.results import (
    ensure_result_set,
    text_encode,
    binary_encode,
    NullBitmap,
)
from mysql_mimic.charset import CharacterSet, Collation
from mysql_mimic import types
from mysql_mimic.utils import seq
from mysql_mimic.admin import Admin

logger = logging.getLogger(__name__)


# Borrowed from mysql-connector-python
REGEX_PARAM = re.compile(r"""\?(?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)""")


class Connection:
    _MAX_PREPARED_STMT_ID = 2**32

    def __init__(
        self,
        stream,
        connection_id,
        session,
        server_capabilities=DEFAULT_SERVER_CAPABILITIES,
        force_cursor=False,
    ):
        """
        Client connection.

        Args:
            stream (mysql_mimic.stream.MysqlStream): stream to use for writing/reading
            connection_id (int): 32 bit connection ID
            session (mysql_mimic.session.Session): session
            server_capabilities (int): server capability flags
            force_cursor (bool): If True, always send a cursor when executing prepared statements,
                even if the client doesn't explicitly request it. This is here to get around a bug
                in mysql-connector-python, which doesn't properly set cursor flags.
        """
        self.stream = stream
        self.session = session
        self.connection_id = connection_id

        self.server_capabilities = server_capabilities
        self.client_capabilities = types.Capabilities(0)
        self.capabilities = types.Capabilities(0)
        self.status_flags = types.ServerStatus(0)

        self.max_packet_size = 0
        self.auth_response = None
        self.client_plugin_name = None
        self.client_connect_attrs = {}
        self.zstd_compression_level = 0
        self.force_cursor = force_cursor

        self.prepared_stmt_seq = seq(self._MAX_PREPARED_STMT_ID)
        self.prepared_stmts = {}

        self.admin = Admin(connection_id=connection_id, session=session)

    @property
    def server_character_set(self):
        return self.admin.server_character_set

    @property
    def client_character_set(self):
        return self.admin.client_character_set

    @property
    def database(self):
        return self.admin.database

    @property
    def username(self):
        return self.admin.username

    async def start(self):
        logger.info("Started new connection: %s", self.connection_id)
        try:
            await self.connection_phase()
            await self.session.init(self)
        except Exception as e:
            await self.stream.write(self.error(e, code=ErrorCode.HANDSHAKE_ERROR))
            raise

        try:
            await self.command_phase()
        finally:
            await self.session.close()

    async def connection_phase(self):
        """https://dev.mysql.com/doc/internals/en/connection-phase.html"""
        await self.stream.write(self.handshake_v10())
        self.handshake_response_41(await self.stream.read())
        await self.stream.write(self.ok())
        self.stream.reset_seq()

    async def command_phase(self):
        """https://dev.mysql.com/doc/internals/en/command-phase.html"""
        while True:
            data = await self.stream.read()
            try:
                command = data[0]
                rest = data[1:]

                if command == types.Commands.COM_QUERY:
                    await self.handle_query(rest)
                elif command == types.Commands.COM_STMT_PREPARE:
                    await self.handle_stmt_prepare(rest)
                elif command == types.Commands.COM_STMT_SEND_LONG_DATA:
                    await self.handle_stmt_send_long_data(rest)
                elif command == types.Commands.COM_STMT_EXECUTE:
                    await self.handle_stmt_execute(rest)
                elif command == types.Commands.COM_STMT_FETCH:
                    await self.handle_stmt_fetch(rest)
                elif command == types.Commands.COM_STMT_RESET:
                    await self.handle_stmt_reset(rest)
                elif command == types.Commands.COM_STMT_CLOSE:
                    await self.handle_stmt_close(rest)
                elif command == types.Commands.COM_PING:
                    await self.handle_ping(rest)
                elif command == types.Commands.COM_RESET_CONNECTION:
                    await self.handle_reset_connection(rest)
                elif command == types.Commands.COM_DEBUG:
                    await self.handle_debug(rest)
                elif command == types.Commands.COM_QUIT:
                    return
                else:
                    raise MysqlError(
                        f"Unsupported Command: {hex(command)}",
                        ErrorCode.UNKNOWN_COM_ERROR,
                    )

            except MysqlError as e:
                logger.exception(e)
                await self.stream.write(self.error(msg=e.msg, code=e.code))
            except Exception as e:  # pylint: disable=broad-except
                logger.exception(e)
                await self.stream.write(self.error(e))
            finally:
                self.stream.reset_seq()

    async def handle_ping(self, data):  # pylint: disable=unused-argument
        """
        https://dev.mysql.com/doc/internals/en/com-ping.html

        COM_PING check if the server is alive.
        """
        await self.stream.write(self.ok())

    async def handle_reset_connection(self, data):  # pylint: disable=unused-argument
        """
        https://dev.mysql.com/doc/internals/en/com-reset-connection.html

        COM_RESET_CONNECTION Resets the session state.

        For now, we're just treating this like a no-op.
        """
        await self.stream.write(self.ok())

    async def handle_debug(self, data):  # pylint: disable=unused-argument
        """
        https://dev.mysql.com/doc/internals/en/com-debug.html

        COM_DEBUG triggers a dump on internal debug info to stdout of the mysql-server.

        For now, we're just treating this like a no-op.
        """
        await self.stream.write(self.ok())

    async def handle_query(self, data):
        """
        https://dev.mysql.com/doc/internals/en/com-query.html

        COM_QUERY is used to send the server a text-based query that is executed immediately.
        """
        sql = data.decode(self.client_character_set.codec)

        result_set = await self.query(sql)

        if not result_set:
            await self.stream.write(self.ok())
            return

        for packet in self.text_resultset(result_set):
            await self.stream.write(packet)

    async def handle_stmt_prepare(self, data):
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-prepare.html

        COM_STMT_PREPARE creates a prepared statement from the passed query string.
        """
        sql = data.decode(self.client_character_set.codec)

        stmt_id = next(self.prepared_stmt_seq)
        num_params = len(REGEX_PARAM.findall(sql))

        stmt = PreparedStatement(
            stmt_id=stmt_id,
            sql=sql,
            num_params=num_params,
        )
        self.prepared_stmts[stmt_id] = stmt

        for packet in self.stmt_prepare_ok(stmt):
            await self.stream.write(packet)

    async def handle_stmt_send_long_data(self, data):
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-send-long-data.html

        COM_STMT_SEND_LONG_DATA sends the data for a column.
        """
        r = io.BytesIO(data)
        stmt_id = types.read_uint_4(r)
        param_id = types.read_uint_2(r)
        stmt = self.get_stmt(stmt_id)
        if stmt.param_buffers is None:
            stmt.param_buffers = {}
        buffer = stmt.param_buffers.setdefault(param_id, bytearray())
        buffer.extend(r.read())

    async def handle_stmt_execute(self, data):
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-execute.html

        COM_STMT_EXECUTE asks the server to execute a prepared statement as identified by stmt-id.
        """
        r = io.BytesIO(data)
        stmt_id = types.read_uint_4(r)
        stmt = self.get_stmt(stmt_id)
        use_cursor = self.read_cursor_flags(r)
        types.read_uint_4(r)  # iteration count
        sql = self.interpolate_params(r, stmt)
        stmt.param_buffers = None

        result_set = await self.query(sql)

        if not result_set:
            await self.stream.write(self.ok())
            return

        await self.stream.write(types.uint_len(len(result_set.columns)))

        for column in result_set.columns:
            await self.stream.write(
                self.column_definition_41(
                    name=column.name,
                    column_type=column.type,
                    character_set=column.character_set,
                )
            )

        rows = (self.binary_resultrow(r) for r in result_set.rows)

        if use_cursor:
            stmt.cursor = rows
            await self.stream.write(
                self.ok_or_eof(flags=types.ServerStatus.SERVER_STATUS_CURSOR_EXISTS)
            )
        else:
            if not self.deprecate_eof():
                self.eof()
            for row in rows:
                await self.stream.write(row)
            await self.stream.write(self.ok_or_eof())

    async def handle_stmt_fetch(self, data):
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-fetch.html

        COM_STMT_FETCH fetches rows from an existing resultset after a COM_STMT_EXECUTE.
        """
        r = io.BytesIO(data)
        stmt_id = types.read_uint_4(r)
        num_rows = types.read_uint_4(r)
        stmt = self.get_stmt(stmt_id)

        count = 0
        for _, packet in zip(range(num_rows), stmt.cursor):
            await self.stream.write(packet)
            count += 1

        done = count < num_rows

        await self.stream.write(
            self.ok_or_eof(
                flags=types.ServerStatus.SERVER_STATUS_LAST_ROW_SENT
                if done
                else types.ServerStatus.SERVER_STATUS_CURSOR_EXISTS
            )
        )

    async def handle_stmt_reset(self, data):
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-reset.html

        COM_STMT_RESET resets the data of a prepared statement which was accumulated with COM_STMT_SEND_LONG_DATA
        commands and closes the cursor if it was opened with COM_STMT_EXECUTE.
        """
        r = io.BytesIO(data)
        stmt_id = types.read_uint_4(r)
        stmt = self.get_stmt(stmt_id)
        stmt.buffers = None
        stmt.cursor = None
        await self.stream.write(self.ok())

    async def handle_stmt_close(self, data):
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-close.html

        COM_STMT_CLOSE deallocates a prepared statement.
        """
        r = io.BytesIO(data)
        stmt_id = types.read_uint_4(r)
        self.prepared_stmts.pop(stmt_id, None)

    def read_cursor_flags(self, reader):
        flags = types.read_uint_1(reader)
        if (
            self.force_cursor
            or flags == types.ComStmtExecuteFlags.CURSOR_TYPE_READ_ONLY
        ):
            return True
        if flags == types.ComStmtExecuteFlags.CURSOR_TYPE_NO_CURSOR:
            return False
        raise MysqlError(
            f"Unsupported cursor flags: {flags}", ErrorCode.NOT_SUPPORTED_YET
        )

    def interpolate_params(self, reader, stmt):
        sql = stmt.sql
        if stmt.num_params:
            null_bitmap = NullBitmap.from_buffer(reader, stmt.num_params)
            new_params_bound_flag = types.read_uint_1(reader)

            if not new_params_bound_flag:
                raise MysqlError(
                    "Server requires the new-params-bound-flag to be set",
                    ErrorCode.NOT_SUPPORTED_YET,
                )

            param_types = []
            for _ in range(stmt.num_params):
                param_type = types.ColumnType(types.read_uint_1(reader))
                is_unsigned = (types.read_uint_1(reader) & 0x80) > 0
                param_types.append((param_type, is_unsigned))

            for param_id, (param_type, is_unsigned) in enumerate(param_types):
                if null_bitmap.is_flipped(param_id):
                    param = "NULL"
                else:
                    param = self.read_param(
                        reader, param_type, is_unsigned, param_id, stmt
                    )
                sql = REGEX_PARAM.sub(param, sql, 1)

        return sql

    def read_param(self, reader, param_type, unsigned, param_id, stmt):
        if stmt.param_buffers and param_id in stmt.param_buffers:
            decoded = bytes(stmt.param_buffers[param_id]).decode(
                self.client_character_set.codec
            )
            return f"'{decoded}'"

        if param_type in {
            types.ColumnType.VARCHAR,
            types.ColumnType.VAR_STRING,
            types.ColumnType.STRING,
            types.ColumnType.BLOB,
            types.ColumnType.TINY_BLOB,
            types.ColumnType.MEDIUM_BLOB,
            types.ColumnType.LONG_BLOB,
        }:
            val = types.read_str_len(reader)
            decoded = val.decode(self.client_character_set.codec)
            return f"'{decoded}'"

        if param_type == types.ColumnType.TINY:
            return str((types.read_uint_1 if unsigned else types.read_int_1)(reader))

        if param_type == types.ColumnType.BOOL:
            return "TRUE" if types.read_uint_1(reader) else "FALSE"

        if param_type in {types.ColumnType.SHORT, types.ColumnType.YEAR}:
            return str((types.read_uint_2 if unsigned else types.read_int_2)(reader))

        if param_type in {types.ColumnType.LONG, types.ColumnType.INT24}:
            return str((types.read_uint_4 if unsigned else types.read_int_4)(reader))

        if param_type == types.ColumnType.LONGLONG:
            return str((types.read_uint_8 if unsigned else types.read_int_8)(reader))

        if param_type == types.ColumnType.FLOAT:
            return str(types.read_float(reader))

        if param_type == types.ColumnType.DOUBLE:
            return str(types.read_double(reader))

        if param_type == types.ColumnType.NULL:
            return "NULL"

        raise MysqlError(
            f"Unsupported parameter type: {param_type}", ErrorCode.NOT_SUPPORTED_YET
        )

    def get_stmt(self, stmt_id):
        if stmt_id in self.prepared_stmts:
            return self.prepared_stmts[stmt_id]
        raise MysqlError(f"Unknown statement: {stmt_id}", ErrorCode.UNKNOWN_PROCEDURE)

    async def query(self, sql):
        result_set = await self.admin.parse(sql)

        if result_set is None:
            sql = self.admin.replace_variables(sql)
            result_set = await self.session.query(sql)
            result_set = result_set and ensure_result_set(result_set)

        return result_set

    def ok(self, eof=False, affected_rows=0, last_insert_id=0, warnings=0, flags=0):
        """https://dev.mysql.com/doc/internals/en/packet-OK_Packet.html"""
        data = types.uint_1(0) if not eof else types.uint_1(0xFE)
        data += types.uint_len(affected_rows) + types.uint_len(last_insert_id)

        if types.Capabilities.CLIENT_PROTOCOL_41 in self.capabilities:
            data += types.uint_2(self.status_flags | flags)
            data += types.uint_2(warnings)
        elif types.Capabilities.CLIENT_TRANSACTIONS in self.capabilities:
            data += types.uint_2(self.status_flags | flags)

        return data

    def eof(self, warnings=0, flags=0):
        """https://dev.mysql.com/doc/internals/en/packet-EOF_Packet.html"""
        data = types.uint_1(0xFE)

        if types.Capabilities.CLIENT_PROTOCOL_41 in self.capabilities:
            data += types.uint_2(warnings)
            data += types.uint_2(self.status_flags | flags)

        return data

    def ok_or_eof(self, affected_rows=0, last_insert_id=0, warnings=0, flags=0):
        if self.deprecate_eof():
            return self.ok(
                eof=True,
                affected_rows=affected_rows,
                last_insert_id=last_insert_id,
                warnings=warnings,
                flags=flags,
            )
        return self.eof(warnings, flags)

    def error(
        self,
        msg,
        code=ErrorCode.UNKNOWN_ERROR,
    ):
        """https://dev.mysql.com/doc/internals/en/packet-ERR_Packet.html"""
        data = types.uint_1(0xFF) + types.uint_2(code)

        if types.Capabilities.CLIENT_PROTOCOL_41 in self.capabilities:
            data += types.str_fixed(1, b"#")
            data += types.str_fixed(5, get_sqlstate(code))

        data += types.str_rest(str(msg).encode(self.server_character_set.codec))

        return data

    # pylint: disable=too-many-arguments
    def column_definition_41(
        self,
        schema=None,
        table=None,
        org_table=None,
        name=None,
        org_name=None,
        character_set=CharacterSet.utf8mb4,
        column_length=256,
        column_type=types.ColumnType.VARCHAR,
        flags=types.ColumnDefinition(0),
        decimals=0,
    ):
        """https://dev.mysql.com/doc/internals/en/com-query-response.html#packet-Protocol::ColumnDefinition41"""
        schema = schema or ""
        table = table or ""
        org_table = org_table or table
        name = name or ""
        org_name = org_name or name

        return (
            types.str_len(b"def")
            + types.str_len(schema.encode(self.server_character_set.codec))
            + types.str_len(table.encode(self.server_character_set.codec))
            + types.str_len(org_table.encode(self.server_character_set.codec))
            + types.str_len(name.encode(self.server_character_set.codec))
            + types.str_len(org_name.encode(self.server_character_set.codec))
            + types.uint_len(0x0C)  # Length of the following fields
            + types.uint_2(character_set)
            + types.uint_4(column_length)
            + types.uint_1(column_type)
            + types.uint_2(flags)
            + types.uint_1(decimals)
            + types.uint_2(0)  # filler
        )

    def handshake_response_41(self, data):
        """https://dev.mysql.com/doc/internals/en/connection-phase-packets.html#packet-Protocol::HandshakeResponse41"""
        r = io.BytesIO(data)
        self.client_capabilities = types.Capabilities(types.read_uint_4(r))
        self.capabilities = self.server_capabilities & self.client_capabilities
        self.max_packet_size = types.read_uint_4(r)
        self.admin.client_character_set = Collation(types.read_uint_1(r)).charset
        types.read_str_fixed(r, 23)
        self.admin.username = types.read_str_null(r).decode()

        if (
            types.Capabilities.CLIENT_PLUGIN_AUTH_LENENC_CLIENT_DATA
            in self.capabilities
        ):
            self.auth_response = types.read_str_len(r).decode()
        else:
            l_auth_response = types.read_uint_1(r)
            self.auth_response = types.read_str_fixed(r, l_auth_response).decode()

        if types.Capabilities.CLIENT_CONNECT_WITH_DB in self.capabilities:
            self.admin.database = types.read_str_null(r).decode()

        if types.Capabilities.CLIENT_PLUGIN_AUTH in self.capabilities:
            self.client_plugin_name = types.read_str_null(r).decode()

        if types.Capabilities.CLIENT_CONNECT_ATTRS in self.capabilities:
            total_l = types.read_uint_len(r)

            while total_l > 0:
                key = types.read_str_len(r)
                value = types.read_str_len(r)
                self.client_connect_attrs[key.decode()] = value.decode()

                item_l = len(types.str_len(key) + types.str_len(value))
                total_l -= item_l

        if types.Capabilities.CLIENT_ZSTD_COMPRESSION_ALGORITHM in self.capabilities:
            self.zstd_compression_level = types.read_uint_1(r)

    def handshake_v10(self):
        """https://dev.mysql.com/doc/internals/en/connection-phase-packets.html#packet-Protocol::HandshakeV10"""
        return (
            types.uint_1(10)  # Always 10
            + types.str_null(
                self.admin.mysql_version.encode(self.server_character_set.codec)
            )  # Status
            + types.uint_4(self.connection_id)  # Connection ID
            + types.str_null(bytes(8))  # plugin data
            + types.uint_2(self.server_capabilities & 0xFFFF)  # lower capabilities flag
            + types.uint_1(self.server_character_set)  # lower character set
            + types.uint_2(self.status_flags)  # server status flag
            + types.uint_2(self.server_capabilities >> 16)  # higher capabilities flag
            + types.uint_1(0)  # constant 0 (no CLIENT_PLUGIN_AUTH)
            + types.str_fixed(10, bytes(10))  # reserved
            + types.str_fixed(13, bytes(13))  # rest of plugin data
        )

    def deprecate_eof(self):
        return types.Capabilities.CLIENT_DEPRECATE_EOF in self.capabilities

    def text_resultset(self, result_set):
        """https://dev.mysql.com/doc/internals/en/com-query-response.html#packet-ProtocolText::Resultset"""
        column_count = b""

        if types.Capabilities.CLIENT_OPTIONAL_RESULTSET_METADATA in self.capabilities:
            column_count += types.uint_1(
                types.ResultsetMetadata.RESULTSET_METADATA_FULL
            )

        column_count += types.uint_len(len(result_set.columns))

        packets = [column_count]

        for column in result_set.columns:
            packets.append(
                self.column_definition_41(
                    name=column.name,
                    column_type=column.type,
                    character_set=column.character_set,
                )
            )

        if not self.deprecate_eof():
            packets.append(self.eof())

        affected_rows = 0

        for row in result_set.rows:
            affected_rows += 1
            row_data = []

            for value, column in zip(row, result_set.columns):
                if value is None:
                    row_data.append(b"\xfb")
                else:
                    text = text_encode(value)
                    if isinstance(text, str):
                        text = text.encode(self.server_character_set.codec)
                    row_data.append(types.str_len(text))
            packets.append(b"".join(row_data))

        packets.append(self.ok_or_eof(affected_rows=affected_rows))

        return packets

    def binary_resultrow(self, row):
        """https://dev.mysql.com/doc/internals/en/binary-protocol-resultset-row.html"""
        column_count = len(row)

        null_bitmap = NullBitmap.new(column_count, offset=2)

        values = []
        for i, val in enumerate(row):
            if val is None:
                null_bitmap.flip(i)
            else:
                values.append(binary_encode(val))

        values_data = b"".join(values)

        null_bitmap_data = bytes(null_bitmap)

        return b"".join(
            [
                types.uint_1(0),  # packet header
                types.str_fixed(len(null_bitmap_data), null_bitmap_data),
                types.str_fixed(len(values_data), values_data),
            ]
        )

    def stmt_prepare_ok(self, statement):
        """https://dev.mysql.com/doc/internals/en/com-stmt-prepare-response.html#packet-COM_STMT_PREPARE_OK"""
        num_params = statement.num_params

        packets = [
            b"".join(
                [
                    types.uint_1(0),  # OK
                    types.uint_4(statement.stmt_id),
                    types.uint_2(0),  # number of columns
                    types.uint_2(num_params),
                    types.uint_1(0),  # filler
                    types.uint_2(0),  # number of warnings
                ]
            )
        ]

        if num_params:
            for _ in range(num_params):
                packets.append(self.column_definition_41(name="?"))
            packets.append(self.eof())

        return packets


@dataclass
class PreparedStatement:
    stmt_id: int
    sql: str
    num_params: int
    param_buffers: Optional[Dict[int, bytearray]] = None
    cursor: Optional[Iterable] = None



import struct
from dataclasses import dataclass
from datetime import datetime, date, timedelta
from typing import Iterable, Sequence, Optional

from mysql_mimic.errors import MysqlError
from mysql_mimic.types import ColumnType, str_len, uint_1, uint_2, uint_4
from mysql_mimic.charset import CharacterSet


@dataclass
class ResultColumn:
    """Column data for a result set"""

    name: str
    type: ColumnType
    character_set: CharacterSet = CharacterSet.utf8mb4


@dataclass
class ResultSet:
    rows: Iterable[Sequence]
    columns: Sequence[ResultColumn]

    def __bool__(self):
        return bool(self.columns)


@dataclass
class Column:
    """
    Column data for a SHOW COLUMNS statement

    https://dev.mysql.com/doc/refman/8.0/en/show-columns.html
    """

    name: str
    type: str
    collation: str = "NULL"
    null: str = "YES"
    key: Optional[str] = None
    default: Optional[str] = None
    extra: Optional[str] = None
    privileges: Optional[str] = None
    comment: Optional[str] = None


def ensure_result_set(result):
    if isinstance(result, ResultSet):
        return result
    if isinstance(result, tuple):
        if len(result) != 2:
            raise MysqlError(
                f"Result tuple should be of size 2. Received: {len(result)}"
            )
        rows = result[0]
        columns = result[1]

        columns = [_ensure_result_col(col, i, rows) for i, col in enumerate(columns)]
        return ResultSet(
            rows=rows,
            columns=columns,
        )

    raise MysqlError(f"Unexpected result set type: {type(result)}")


def _ensure_result_col(column, idx, rows):
    if isinstance(column, ResultColumn):
        return column

    if isinstance(column, str):
        value = _find_first_non_null_value(idx, rows)
        type_ = infer_type(value)
        return ResultColumn(
            name=column,
            type=type_,
        )

    raise MysqlError(f"Unexpected result column value: {column}")


def _find_first_non_null_value(idx, rows):
    for row in rows:
        value = row[idx]
        if value is not None:
            return value
    return None


def _binary_encode_bool(val):
    return uint_1(int(val))


def _binary_encode_str(val):
    return str_len(str(val).encode("utf-8"))


def _binary_encode_bytes(val):
    return str_len(val)


def _binary_encode_date(val):
    year = val.year
    month = val.month
    day = val.day

    if isinstance(val, datetime):
        hour = val.hour
        minute = val.minute
        second = val.second
        microsecond = val.microsecond
    else:
        hour = minute = second = microsecond = 0

    if microsecond == 0:
        if hour == minute == second == 0:
            if year == month == day == 0:
                return uint_1(0)
            return b"".join([uint_1(4), uint_2(year), uint_1(month), uint_1(day)])
        return b"".join(
            [
                uint_1(7),
                uint_2(year),
                uint_1(month),
                uint_1(day),
                uint_1(hour),
                uint_1(minute),
                uint_1(second),
            ]
        )
    return b"".join(
        [
            uint_1(11),
            uint_2(year),
            uint_1(month),
            uint_1(day),
            uint_1(hour),
            uint_1(minute),
            uint_1(second),
            uint_4(microsecond),
        ]
    )


def _binary_encode_int(val):
    return struct.pack("<q", val)


def _binary_encode_float(val):
    return struct.pack("<d", val)


def _binary_encode_timedelta(val):
    days = abs(val.days)
    hours, remainder = divmod(abs(val.seconds), 3600)
    minutes, seconds = divmod(remainder, 60)
    microseconds = val.microseconds
    is_negative = val.total_seconds() < 0

    if microseconds == 0:
        if days == hours == minutes == seconds == 0:
            return uint_1(0)
        return b"".join(
            [
                uint_1(8),
                uint_1(is_negative),
                uint_4(days),
                uint_1(hours),
                uint_1(minutes),
                uint_1(seconds),
            ]
        )
    return b"".join(
        [
            uint_1(12),
            uint_1(is_negative),
            uint_4(days),
            uint_1(hours),
            uint_1(minutes),
            uint_1(seconds),
            uint_4(microseconds),
        ]
    )


# Order matters
# bool is a subclass of int
# datetime is a subclass of date
_ENCODERS = {
    # python type: (mysql type, text encoder, binary encoder),
    bool: (ColumnType.TINY, lambda v: str(int(v)), _binary_encode_bool),
    datetime: (ColumnType.DATETIME, str, _binary_encode_date),
    str: (ColumnType.STRING, lambda v: v, _binary_encode_str),
    bytes: (ColumnType.BLOB, lambda v: v, _binary_encode_bytes),
    int: (ColumnType.LONGLONG, str, _binary_encode_int),
    float: (ColumnType.DOUBLE, str, _binary_encode_float),
    date: (ColumnType.DATE, str, _binary_encode_date),
    timedelta: (ColumnType.TIME, str, _binary_encode_timedelta),
}


def binary_encode(val):
    for py_type, (_, _, encoder) in _ENCODERS.items():
        if isinstance(val, py_type):
            return encoder(val)
    return _binary_encode_str(val)


def text_encode(val):
    for py_type, (_, encoder, _) in _ENCODERS.items():
        if isinstance(val, py_type):
            return encoder(val)
    return str(val)


def infer_type(val):
    for py_type, (my_type, _, _) in _ENCODERS.items():
        if isinstance(val, py_type):
            return my_type
    return ColumnType.VARCHAR


class NullBitmap:
    """See https://dev.mysql.com/doc/internals/en/null-bitmap.html"""

    __slots__ = ("offset", "bitmap")

    def __init__(self, bitmap, offset=0):
        self.offset = offset
        self.bitmap = bitmap

    @classmethod
    def new(cls, num_bits, offset=0):
        bitmap = bytearray(cls._num_bytes(num_bits, offset))
        return cls(bitmap, offset)

    @classmethod
    def from_buffer(cls, buffer, num_bits, offset=0):
        bitmap = bytearray(buffer.read(cls._num_bytes(num_bits, offset)))
        return cls(bitmap, offset)

    @classmethod
    def _num_bytes(cls, num_bits, offset):
        return (num_bits + 7 + offset) // 8

    def flip(self, i):
        byte_position, bit_position = self._pos(i)
        self.bitmap[byte_position] |= 1 << bit_position

    def is_flipped(self, i):
        byte_position, bit_position = self._pos(i)
        return bool(self.bitmap[byte_position] & (1 << bit_position))

    def _pos(self, i):
        byte_position = (i + self.offset) // 8
        bit_position = (i + self.offset) % 8
        return byte_position, bit_position

    def __bytes__(self):
        return bytes(self.bitmap)



from collections.abc import Iterator


class seq(Iterator):
    """Auto-incrementing sequence with an optional maximum size"""

    def __init__(self, size=None):
        self.size = size
        self.value = 0

    def __next__(self):
        value = self.value
        self.value = self.value + 1
        if self.size:
            self.value = self.value % self.size
        return value

    def reset(self):
        self.value = 0



import re

from mysql_mimic.charset import CharacterSet, Collation
from mysql_mimic.errors import MysqlError, ErrorCode
from mysql_mimic.results import ResultColumn, ResultSet, Column
from mysql_mimic.types import ColumnType


class Admin:
    """
    Administration Statements (plus some other things)
    https://dev.mysql.com/doc/refman/8.0/en/sql-server-administration-statements.html

    Args:
        connection_id (int): connection ID
        session (mysql_mimic.session.Session): session
        variable_defaults (dict): default values for session variables
    """

    # Regex for finding "information functions" to replace in SQL statements
    REGEX_INFO_FUNC = re.compile(
        r"""
            (?P<func>
                (CONNECTION_ID)
                |(USER)
                |(SYSTEM_USER)
                |(SESSION_USER)
                |(VERSION)
            )\(\)
            (?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)
        """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for finding "session variables" to replace in SQL statements
    REGEX_SESSION_VAR = re.compile(
        r"""
            @@(?P<var>\w+)
            (?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)
        """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for determining if a statement is an "admin statement"
    REGEX_CMD = re.compile(
        r"^\s*(?P<cmd>(SET)|(SHOW))\s(?P<this>.*)$", re.IGNORECASE | re.VERBOSE
    )

    # Regex for parsing a "SET VARIABLES" statement
    # Not yet supported:
    # 1. Setting variable to result of scalar subquery
    # 2. Setting variable to another variable
    # 3. Setting multiple variables at once
    REGEX_SET_VAR = re.compile(
        r"""
        ^\s*
        (
            (?P<global>(GLOBAL\s+)|(@@GLOBAL\.))
            |(?P<persist>(PERSIST\s+)|(@@PERSIST\.))
            |(?P<persist_only>(PERSIST_ONLY\s+)|(@@PERSIST_ONLY\.))
            |(?P<session>(SESSION\s+)|(@@SESSION\.)|(@@))
            |(?P<user>@)
        )?
        (?P<var_name>(\w+))
        \s*
        =
        \s*(?P<value>.*?)
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SET NAMES" statement
    REGEX_SET_NAMES = re.compile(
        r"""
        ^\s*
        NAMES\s+
        '?(?P<charset_name>\w+)'?
        \s*
        (
            (COLLATE\s+'?(?P<collation_name>\w+)'?)
            |
            (DEFAULT)
        )?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW VARIABLES" statement
    REGEX_SHOW_VARS = re.compile(
        r"""
        ^\s*
        (
            (?P<global>GLOBAL)
            |(?P<session>SESSION)
            \s*
        )?
        VARIABLES\s*
        (LIKE\s*'(?P<like>[\w%]+)'\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW COLUMNS" statement
    REGEX_SHOW_COLUMNS = re.compile(
        r"""
        ^\s*
        ((?P<extended>EXTENDED)\s*)?
        ((?P<full>FULL)\s*)?
        ((COLUMNS)|(FIELDS))\s*
        ((FROM)|(IN))\s*(`?(?P<impl_db_name>\w+)`?\.)?`?(?P<tbl_name>\w+)`?\s*
        (((FROM)|(IN))\s*`?(?P<expl_db_name>\w+)`?\s*)?
        (LIKE\s*'(?P<like>[\w%]+)'\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW INDEX" statement
    REGEX_SHOW_INDEX = re.compile(
        r"""
        ^\s*
        ((?P<extended>EXTENDED)\s*)?
        ((?P<full>FULL)\s*)?
        ((INDEX)|(INDEXES)|(KEYS))\s*
        ((FROM)|(IN))\s*(`?(?P<impl_db_name>\w+)`?\.)?`?(?P<tbl_name>\w+)`?\s*
        (((FROM)|(IN))\s*`?(?P<expl_db_name>\w+)`?\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    def __init__(self, connection_id, session, variable_defaults=None):
        self.connection_id = connection_id
        self.session = session
        self.database = None
        self.username = None
        self.variable_defaults = variable_defaults or {
            "version": "8.0.29",
            "version_comment": "mysql-mimic",
            "character_set_client": CharacterSet.utf8mb4.name,
            "character_set_results": CharacterSet.utf8mb4.name,
            "character_set_server": CharacterSet.utf8mb4.name,
            "collation_server": Collation.utf8mb4_general_ci.name,
            "collation_database": Collation.utf8mb4_general_ci.name,
        }
        self.variables = dict(**self.variable_defaults)

    @property
    def server_character_set(self):
        return CharacterSet[self.variables.get("character_set_results", "utf8mb4")]

    @server_character_set.setter
    def server_character_set(self, val):
        self.variables["character_set_results"] = val.name

    @property
    def client_character_set(self):
        return CharacterSet[self.variables.get("character_set_client", "utf8mb4")]

    @client_character_set.setter
    def client_character_set(self, val):
        self.variables["character_set_client"] = val.name

    @property
    def mysql_version(self):
        return self.variables["version"]

    def replace_variables(self, sql):
        sql = self.REGEX_INFO_FUNC.sub(self._replace_info_func, sql)
        return self.REGEX_SESSION_VAR.sub(self._replace_session_var, sql)

    def _replace_info_func(self, matchobj):
        func = matchobj.group("func").lower()
        if func == "connection_id":
            return str(self.connection_id)
        if func in {"user", "session_user", "system_user"}:
            return f"'{self.username}'"
        if func == "version":
            return f"'{self.mysql_version}'"
        raise MysqlError(f"Failed to parse system information function: {func}")

    def _replace_session_var(self, matchobj):
        var = matchobj.group("var").lower()
        if var in self.variables:
            return f"'{self.variables[var]}'"
        raise MysqlError(f"Unknown variable: {var}", ErrorCode.UNKNOWN_SYSTEM_VARIABLE)

    async def parse(self, sql):
        m = self.REGEX_CMD.match(sql)
        if not m:
            return None

        cmd = m.group("cmd").lower()
        this = m.group("this").lower()

        if cmd == "set":
            return await self._parse_set(this)
        if cmd == "show":
            return await self._parse_show(this)
        raise MysqlError("Failed to parse command", ErrorCode.PARSE_ERROR)

    async def _parse_set(self, this):
        m = self.REGEX_SET_NAMES.match(this)
        if m:
            return await self._parse_set_names(
                charset_name=m.group("charset_name"),
                collation_name=m.group("collation_name"),
            )

        m = self.REGEX_SET_VAR.match(this)
        if m:
            return await self._parse_set_variables(
                global_=m.group("global"),
                persist=m.group("persist"),
                persist_only=m.group("persist_only"),
                session=m.group("session"),
                user=m.group("user"),
                var_name=m.group("var_name"),
                value=m.group("value"),
            )

        raise MysqlError("Unsupported SET command", ErrorCode.NOT_SUPPORTED_YET)

    async def _set_variable(self, key, val):
        self.variables[key] = val
        await self.session.set(**{key: val})

    async def _parse_set_names(self, charset_name, collation_name):
        await self._set_variable("character_set_client", charset_name)
        await self._set_variable("character_set_connection", charset_name)
        await self._set_variable("character_set_results", charset_name)
        await self._set_variable(
            "collation_connection",
            (collation_name or CharacterSet[charset_name].default_collation.name),
        )
        return ResultSet([], [])

    async def _parse_set_variables(
        self, global_, persist, persist_only, session, user, var_name, value
    ):  # pylint: disable=unused-argument
        if global_ or persist or persist_only or user:
            raise MysqlError(
                "Only setting session variables is supported",
                ErrorCode.NOT_SUPPORTED_YET,
            )

        value_lower = value.lower()
        if value_lower in {"true", "on"}:
            result = True
        elif value_lower in {"false", "off"}:
            result = False
        elif value_lower in {"default", "null"}:
            result = self.variable_defaults.get(var_name)
        elif value[0] == value[-1] == "'":
            result = value[1:-1]
        else:
            try:
                result = int(value)
            except ValueError:
                try:
                    result = float(value)
                except ValueError as e:
                    raise MysqlError(
                        f"Unexpected variable value: {value}",
                        ErrorCode.WRONG_VALUE_FOR_VAR,
                    ) from e

        await self._set_variable(var_name, result)
        return ResultSet([], [])

    async def _parse_show(self, this):
        m = self.REGEX_SHOW_VARS.match(this)
        if m:
            return await self._parse_show_variables(
                global_=bool(m.group("global")),
                session=bool(m.group("session")),
                like=self._like_to_regex(m.group("like")),
            )
        m = self.REGEX_SHOW_COLUMNS.match(this)
        if m:
            return await self._parse_show_columns(
                extended=bool(m.group("extended")),
                full=bool(m.group("full")),
                db_name=m.group("expl_db_name") or m.group("impl_db_name"),
                tbl_name=m.group("tbl_name"),
                like=self._like_to_regex(m.group("like")),
            )

        m = self.REGEX_SHOW_INDEX.match(this)
        if m:
            return await self._parse_show_index(
                extended=bool(m.group("extended")),
                db_name=m.group("expl_db_name") or m.group("impl_db_name"),
                tbl_name=m.group("tbl_name"),
            )

        raise MysqlError("Unsupported SHOW command", ErrorCode.NOT_SUPPORTED_YET)

    def _like_to_regex(self, like):
        if like is None:
            return re.compile(r".*")
        like = like.replace("%", ".*")
        like = like.replace("_", ".")
        return re.compile(like)

    async def _parse_show_columns(
        self, extended, full, db_name, tbl_name, like
    ):  # pylint: disable=unused-argument
        db_name = db_name or self.database
        if not db_name:
            raise MysqlError("No database selected", ErrorCode.NO_DB_ERROR)

        columns = await self.session.show_columns(db_name, tbl_name)
        columns = [c if isinstance(c, Column) else Column(**c) for c in columns]
        columns = [c for c in columns if like.match(c.name)]

        result_columns = [
            ResultColumn("Field", ColumnType.STRING),
            ResultColumn("Type", ColumnType.STRING),
            ResultColumn("Null", ColumnType.STRING),
            ResultColumn("Key", ColumnType.STRING),
            ResultColumn("Default", ColumnType.STRING),
            ResultColumn("Extra", ColumnType.STRING),
        ]
        if full:
            result_columns.extend(
                [
                    ResultColumn("Collation", ColumnType.STRING),
                    ResultColumn("Privileges", ColumnType.STRING),
                    ResultColumn("Comment", ColumnType.STRING),
                ]
            )
        rows = []
        for c in columns:
            row = [c.name, c.type, c.null, c.key, c.default, c.extra]
            if full:
                row.extend(
                    [
                        c.collation,
                        c.privileges,
                        c.comment,
                    ]
                )
            rows.append(row)
        return ResultSet(rows=rows, columns=result_columns)

    async def _parse_show_index(
        self, extended, db_name, tbl_name
    ):  # pylint: disable=unused-argument
        result_columns = [
            ResultColumn("Table", ColumnType.STRING),
            ResultColumn("Non_unique", ColumnType.TINY),
            ResultColumn("Key_name", ColumnType.STRING),
            ResultColumn("Seq_in_index", ColumnType.LONGLONG),
            ResultColumn("Column_name", ColumnType.STRING),
            ResultColumn("Collation", ColumnType.STRING),
            ResultColumn("Cardinality", ColumnType.LONGLONG),
            ResultColumn("Sub_part", ColumnType.STRING),
            ResultColumn("Packed", ColumnType.STRING),
            ResultColumn("Null", ColumnType.STRING),
            ResultColumn("Index_type", ColumnType.STRING),
            ResultColumn("Comment", ColumnType.STRING),
            ResultColumn("Index_comment", ColumnType.STRING),
            ResultColumn("Visible", ColumnType.STRING),
            ResultColumn("Expression", ColumnType.STRING),
        ]
        return ResultSet(rows=[], columns=result_columns)

    async def _parse_show_variables(
        self, global_, session, like
    ):  # pylint: disable=unused-argument
        rows = list(self.variables.items())
        rows = [(k, v) for k, v in rows if like.match(k)]
        return ResultSet(
            rows=rows,
            columns=[
                ResultColumn("Variable_name", ColumnType.STRING),
                ResultColumn("Value", ColumnType.STRING),
            ],
        )



import io
import struct
import unittest

from mysql_mimic import types


class TestTypes(unittest.TestCase):
    def test_column_definition(self):
        self.assertEqual(types.ColumnDefinition.NOT_NULL_FLAG, 1)
        self.assertEqual(types.ColumnDefinition.PRI_KEY_FLAG, 2)
        self.assertEqual(types.ColumnDefinition.FIELD_IS_INVISIBLE, 1 << 30)
        self.assertEqual(
            types.ColumnDefinition.NOT_NULL_FLAG | types.ColumnDefinition.PRI_KEY_FLAG,
            3,
        )

    def test_capabilities(self):
        self.assertEqual(types.Capabilities.CLIENT_LONG_PASSWORD, 1)
        self.assertEqual(types.Capabilities.CLIENT_FOUND_ROWS, 2)
        self.assertEqual(types.Capabilities.CLIENT_PROTOCOL_41, 512)
        self.assertEqual(types.Capabilities.CLIENT_DEPRECATE_EOF, 1 << 24)
        self.assertEqual(
            types.Capabilities.CLIENT_LONG_PASSWORD
            | types.Capabilities.CLIENT_FOUND_ROWS,
            3,
        )

    def test_server_status(self):
        self.assertEqual(types.ServerStatus.SERVER_STATUS_IN_TRANS, 1)
        self.assertEqual(types.ServerStatus.SERVER_STATUS_AUTOCOMMIT, 2)
        self.assertEqual(types.ServerStatus.SERVER_STATUS_LAST_ROW_SENT, 128)
        self.assertEqual(
            types.ServerStatus.SERVER_STATUS_IN_TRANS
            | types.ServerStatus.SERVER_STATUS_AUTOCOMMIT,
            3,
        )

    def test_int_1(self):
        self.assertEqual(types.uint_1(0), b"\x00")
        self.assertEqual(types.uint_1(1), b"\x01")
        self.assertEqual(types.uint_1(255), b"\xff")
        self.assertRaises(struct.error, types.uint_1, 256)

    def test_int_2(self):
        self.assertEqual(types.uint_2(0), b"\x00\x00")
        self.assertEqual(types.uint_2(1), b"\x01\x00")
        self.assertEqual(types.uint_2(255), b"\xff\x00")
        self.assertEqual(types.uint_2(2**8), b"\x00\x01")
        self.assertEqual(types.uint_2(2**16 - 1), b"\xff\xff")
        self.assertRaises(struct.error, types.uint_2, 2**16)

    def test_int_3(self):
        self.assertEqual(types.uint_3(0), b"\x00\x00\x00")
        self.assertEqual(types.uint_3(1), b"\x01\x00\x00")
        self.assertEqual(types.uint_3(255), b"\xff\x00\x00")
        self.assertEqual(types.uint_3(2**8), b"\x00\x01\x00")
        self.assertEqual(types.uint_3(2**16), b"\x00\x00\x01")
        self.assertRaises(struct.error, types.uint_3, 2**24)

    def test_int_4(self):
        self.assertEqual(types.uint_4(0), b"\x00\x00\x00\x00")
        self.assertEqual(types.uint_4(1), b"\x01\x00\x00\x00")
        self.assertEqual(types.uint_4(255), b"\xff\x00\x00\x00")
        self.assertEqual(types.uint_4(2**8), b"\x00\x01\x00\x00")
        self.assertEqual(types.uint_4(2**16), b"\x00\x00\x01\x00")
        self.assertEqual(types.uint_4(2**24), b"\x00\x00\x00\x01")
        self.assertRaises(struct.error, types.uint_4, 2**32)

    def test_int_6(self):
        self.assertEqual(types.uint_6(0), b"\x00\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_6(1), b"\x01\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_6(255), b"\xff\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_6(2**8), b"\x00\x01\x00\x00\x00\x00")
        self.assertEqual(types.uint_6(2**16), b"\x00\x00\x01\x00\x00\x00")
        self.assertEqual(types.uint_6(2**24), b"\x00\x00\x00\x01\x00\x00")
        self.assertEqual(types.uint_6(2**32), b"\x00\x00\x00\x00\x01\x00")
        self.assertEqual(types.uint_6(2**40), b"\x00\x00\x00\x00\x00\x01")
        self.assertRaises(struct.error, types.uint_6, 2**48)

    def test_int_8(self):
        self.assertEqual(types.uint_8(0), b"\x00\x00\x00\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_8(1), b"\x01\x00\x00\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_8(255), b"\xff\x00\x00\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_8(2**8), b"\x00\x01\x00\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_8(2**16), b"\x00\x00\x01\x00\x00\x00\x00\x00")
        self.assertEqual(types.uint_8(2**24), b"\x00\x00\x00\x01\x00\x00\x00\x00")
        self.assertEqual(types.uint_8(2**32), b"\x00\x00\x00\x00\x01\x00\x00\x00")
        self.assertEqual(types.uint_8(2**56), b"\x00\x00\x00\x00\x00\x00\x00\x01")
        self.assertRaises(struct.error, types.uint_8, 2**64)

    def test_int_len(self):
        self.assertEqual(types.uint_len(0), b"\x00")
        self.assertEqual(types.uint_len(1), b"\x01")
        self.assertEqual(types.uint_len(250), b"\xfa")
        self.assertEqual(types.uint_len(251), b"\xfc\xfb\x00")
        self.assertEqual(types.uint_len(2**8), b"\xfc\x00\x01")
        self.assertEqual(types.uint_len(2**16), b"\xfd\x00\x00\x01")
        self.assertEqual(
            types.uint_len(2**24), b"\xfe\x00\x00\x00\x01\x00\x00\x00\x00"
        )
        self.assertEqual(
            types.uint_len(2**32), b"\xfe\x00\x00\x00\x00\x01\x00\x00\x00"
        )
        self.assertEqual(
            types.uint_len(2**56), b"\xfe\x00\x00\x00\x00\x00\x00\x00\x01"
        )
        self.assertRaises(struct.error, types.uint_len, 2**64)

    def test_str_fixed(self):
        self.assertEqual(types.str_fixed(0, b"kelsin"), b"")
        self.assertEqual(types.str_fixed(1, b"kelsin"), b"k")
        self.assertEqual(types.str_fixed(6, b"kelsin"), b"kelsin")
        self.assertEqual(types.str_fixed(8, b"kelsin"), b"kelsin\x00\x00")

    def test_str_null(self):
        self.assertEqual(types.str_null(b""), b"\x00")
        self.assertEqual(types.str_null(b"kelsin"), b"kelsin\x00")

    def test_str_len(self):
        self.assertEqual(types.str_len(b""), b"\x00")
        self.assertEqual(types.str_len(b"kelsin"), b"\x06kelsin")

        big_str = bytes(256)
        self.assertEqual(types.str_len(big_str), b"\xfc\x00\x01" + big_str)

    def test_str_rest(self):
        self.assertEqual(types.str_rest(b""), b"")
        self.assertEqual(types.str_rest(b"kelsin"), b"kelsin")

        big_str = bytes(256)
        self.assertEqual(types.str_rest(big_str), big_str)

    def test_read_int_1(self):
        reader = io.BytesIO(b"\x01")
        self.assertEqual(types.read_uint_1(reader), 1)
        reader = io.BytesIO(b"\x00\x00")
        self.assertEqual(types.read_uint_1(reader), 0)
        self.assertEqual(types.read_uint_1(reader), 0)

    def test_read_int_2(self):
        reader = io.BytesIO(b"\x00\x01")
        self.assertEqual(types.read_uint_2(reader), 256)

    def test_read_int_3(self):
        reader = io.BytesIO(b"\x00\x00\x01")
        self.assertEqual(types.read_uint_3(reader), 2**16)

    def test_read_int_4(self):
        reader = io.BytesIO(b"\x01\x00\x00\x00")
        self.assertEqual(types.read_uint_4(reader), 1)
        reader = io.BytesIO(b"\x00\x00\x01\x00")
        self.assertEqual(types.read_uint_4(reader), 2**16)
        reader = io.BytesIO(b"\x00\x00\x00\x01")
        self.assertEqual(types.read_uint_4(reader), 2**24)

    def test_read_int_6(self):
        reader = io.BytesIO(b"\x01\x00\x00\x00\x00\x00")
        self.assertEqual(types.read_uint_6(reader), 1)
        reader = io.BytesIO(b"\x00\x00\x00\x00\x01\x00")
        self.assertEqual(types.read_uint_6(reader), 2**32)
        reader = io.BytesIO(b"\x00\x00\x00\x00\x00\x01")
        self.assertEqual(types.read_uint_6(reader), 2**40)

    def test_read_int_8(self):
        reader = io.BytesIO(b"\x01\x00\x00\x00\x00\x00\x00\x00")
        self.assertEqual(types.read_uint_8(reader), 1)
        reader = io.BytesIO(b"\x00\x00\x00\x00\x00\x00\x01\x00")
        self.assertEqual(types.read_uint_8(reader), 2**48)
        reader = io.BytesIO(b"\x00\x00\x00\x00\x00\x00\x00\x01")
        self.assertEqual(types.read_uint_8(reader), 2**56)

    def test_read_int_len(self):
        reader = io.BytesIO(b"\x01")
        self.assertEqual(types.read_uint_len(reader), 1)
        reader = io.BytesIO(b"\xfa")
        self.assertEqual(types.read_uint_len(reader), 250)
        reader = io.BytesIO(b"\xfc\x00\x01")
        self.assertEqual(types.read_uint_len(reader), 256)
        reader = io.BytesIO(b"\xfd\x00\x00\x01")
        self.assertEqual(types.read_uint_len(reader), 2**16)
        reader = io.BytesIO(b"\xfe\x00\x00\x00\x00\x00\x00\x00\x01")
        self.assertEqual(types.read_uint_len(reader), 2**56)

    def test_read_str_fixed(self):
        reader = io.BytesIO(b"kelsin")
        self.assertEqual(types.read_str_fixed(reader, 1), b"k")
        self.assertEqual(types.read_str_fixed(reader, 2), b"el")
        self.assertEqual(types.read_str_fixed(reader, 3), b"sin")

    def test_read_str_null(self):
        reader = io.BytesIO(b"\x00")
        self.assertEqual(types.read_str_null(reader), b"")
        reader = io.BytesIO(b"kelsin\x00")
        self.assertEqual(types.read_str_null(reader), b"kelsin")
        reader = io.BytesIO(b"kelsin\x00foo")
        self.assertEqual(types.read_str_null(reader), b"kelsin")

    def test_read_str_len(self):
        reader = io.BytesIO(b"\x01kelsin")
        self.assertEqual(types.read_str_len(reader), b"k")

        big_str = bytes(256)
        reader = io.BytesIO(b"\xfc\x00\x01" + big_str)
        self.assertEqual(types.read_str_len(reader), big_str)

    def test_read_str_rest(self):
        reader = io.BytesIO(b"kelsin")
        self.assertEqual(types.read_str_rest(reader), b"kelsin")
        reader = io.BytesIO(b"kelsin\x00foo")
        self.assertEqual(types.read_str_rest(reader), b"kelsin\x00foo")



import asyncio
import functools
import io
import unittest
import socket
from datetime import date, datetime, timedelta
from functools import partial

import aiomysql
import mysql.connector
from mysql.connector.connection import MySQLCursorPrepared, MySQLCursorDict

from mysql_mimic import MysqlServer, Session
from mysql_mimic.constants import DEFAULT_SERVER_CAPABILITIES
from mysql_mimic.results import ResultColumn, ResultSet
from mysql_mimic.types import Capabilities, ColumnType
from mysql_mimic.charset import CharacterSet


class PreparedDictCursor(MySQLCursorPrepared):
    def fetchall(self):
        rows = super().fetchall()

        if rows is not None:
            return [dict(zip(self.column_names, row)) for row in rows]

        return None


class MockSession(Session):
    def __init__(self):
        super().__init__()
        self.return_value = None
        self.echo = False
        self.connection = None

    async def init(self, connection):
        self.connection = connection

    async def query(self, sql):
        if self.echo:
            return [(sql,)], ["sql"]
        return self.return_value


async def to_thread(func, *args, **kwargs):
    loop = asyncio.get_running_loop()
    func_call = functools.partial(func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)


def get_free_port():
    sock = socket.socket()
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


class TestIntegration(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        # pylint: disable=attribute-defined-outside-init
        self.session = MockSession()

        self.port = get_free_port()
        self.server = MysqlServer(
            session_factory=lambda: self.session,
            capabilities=(
                DEFAULT_SERVER_CAPABILITIES
                | Capabilities.CLIENT_CONNECT_WITH_DB
                | Capabilities.CLIENT_CONNECT_ATTRS
            ),
            port=self.port,
            conn_kwargs={"force_cursor": True},
        )

        await self.server.start_server()
        asyncio.create_task(self.server.serve_forever())
        self.mysql_conn = await to_thread(
            mysql.connector.connect, use_pure=True, port=self.port
        )
        self.aiomysql_conn = await aiomysql.connect(port=self.port)

    async def asyncTearDown(self):
        self.mysql_conn.close()
        await self.aiomysql_conn.ensure_closed()
        self.server.close()
        await self.server.wait_closed()

    async def mysql_query(self, sql, cursor_class=MySQLCursorDict, params=None):
        cursor = await to_thread(self.mysql_conn.cursor, cursor_class=cursor_class)
        await to_thread(cursor.execute, sql, *(p for p in [params] if p))
        result = await to_thread(cursor.fetchall)
        await to_thread(cursor.close)
        return result

    async def aiomysql_query(self, sql, cursor_class=aiomysql.DictCursor, params=None):
        async with self.aiomysql_conn.cursor(cursor_class) as cur:
            await cur.execute(sql, *(p for p in [params] if p))
            return await cur.fetchall()

    async def test_query(self):
        for query in [
            self.mysql_query,
            partial(self.mysql_query, cursor_class=PreparedDictCursor),
            self.aiomysql_query,
        ]:
            for rv, expected in [
                (None, []),
                (((("x",),), ("b",)), [{"b": "x"}]),
                (([["1"]], ["b"]), [{"b": "1"}]),
                (([(1,)], ("b",)), [{"b": 1}]),
                (([(1.0,)], ("b",)), [{"b": 1.0}]),
                (([(b"x",)], ("b",)), [{"b": "x"}]),
                (([(True,)], ("b",)), [{"b": True}]),
                (([(date(2021, 1, 1),)], ("b",)), [{"b": date(2021, 1, 1)}]),
                (
                    ([(datetime(2021, 1, 1, 1, 1, 1),)], ("b",)),
                    [{"b": datetime(2021, 1, 1, 1, 1, 1)}],
                ),
                (([(timedelta(seconds=60),)], ("b",)), [{"b": timedelta(seconds=60)}]),
                (
                    ResultSet(
                        rows=[("♥".encode("utf-8"),)],
                        columns=[
                            ResultColumn(
                                name="b",
                                character_set=CharacterSet.utf8mb4,
                                type=ColumnType.VARCHAR,
                            )
                        ],
                    ),
                    [{"b": "♥"}],
                ),
                (
                    (
                        [("♥".encode("utf-8"),)],
                        [
                            ResultColumn(
                                name="b",
                                character_set=CharacterSet.utf8mb4,
                                type=ColumnType.VARCHAR,
                            )
                        ],
                    ),
                    [{"b": "♥"}],
                ),
                (([(None,)], ["b"]), [{"b": None}]),
                (([(None, 1, 1)], ["a", "b", "c"]), [{"a": None, "b": 1, "c": 1}]),
                (([(1, None, 1)], ["a", "b", "c"]), [{"a": 1, "b": None, "c": 1}]),
                (([(1, 1, None)], ["a", "b", "c"]), [{"a": 1, "b": 1, "c": None}]),
                (([[None], [1]], ["b"]), [{"b": None}, {"b": 1}]),
            ]:
                self.session.return_value = rv
                result = await query("SELECT b FROM a")
                self.assertEqual(result, expected)

    async def test_prepared_stmt(self):
        self.session.echo = True

        for sql, params, expected in [
            ("SELECT ?", ("1",), "SELECT '1'"),
            ("SELECT '?'", (), "SELECT '?'"),
            ("SELECT ?", (1,), "SELECT 1"),
            ("SELECT ?", (0,), "SELECT 0"),
            ("SELECT ?", (-1,), "SELECT -1"),
            ("SELECT ?", (255,), "SELECT 255"),
            ("SELECT ?", (-128,), "SELECT -128"),
            ("SELECT ?", (65535,), "SELECT 65535"),
            ("SELECT ?", (-32767,), "SELECT -32767"),
            ("SELECT ?", (4294967295,), "SELECT 4294967295"),
            ("SELECT ?", (-2147483648,), "SELECT -2147483648"),
            ("SELECT ?", (18446744073709551615,), "SELECT 18446744073709551615"),
            ("SELECT ?", (-9223372036854775808,), "SELECT -9223372036854775808"),
            ("SELECT ?", (1.1,), "SELECT 1.1"),
            ("SELECT ?", (1.7e308,), "SELECT 1.7e+308"),
            ("SELECT ?", (None,), "SELECT NULL"),
            ("SELECT ?", (b"hello",), "SELECT 'hello'"),
            ("SELECT ?", (io.BytesIO(b"hello"),), "SELECT 'hello'"),
            ("SELECT ?, ?", ("1", "1"), "SELECT '1', '1'"),
            (
                "SELECT ?, ?, ?, ?",
                ("1", None, io.BytesIO(b"hello"), 1),
                "SELECT '1', NULL, 'hello', 1",
            ),
        ]:
            result = await self.mysql_query(
                sql, cursor_class=PreparedDictCursor, params=params
            )
            self.assertEqual(result[0]["sql"], expected)

    async def test_init(self):
        async with aiomysql.connect(
            port=self.port, user="levon_helm", db="db", program_name="test"
        ):
            connection = self.session.connection
            self.assertEqual(connection.username, "levon_helm")
            self.assertEqual(connection.client_connect_attrs["program_name"], "test")
            self.assertEqual(connection.database, "db")

    async def test_connection_id(self):
        async with aiomysql.connect(port=self.port) as conn1:
            async with aiomysql.connect(port=self.port) as conn2:
                self.assertEqual(
                    conn1.server_thread_id[0] + 1, conn2.server_thread_id[0]
                )

    async def test_replace_variables(self):
        self.session.echo = True

        result = await self.aiomysql_query("SELECT CONNECTION_ID()")
        parts = result[0]["sql"].split(" ")
        self.assertEqual(parts[0], "SELECT")
        self.assertTrue(int(parts[1]))

        result = await self.aiomysql_query("SELECT @@version_comment")
        parts = result[0]["sql"].split(" ")
        self.assertEqual(parts[0], "SELECT")
        self.assertEqual(parts[1], "'mysql-mimic'")



import unittest

from mysql_mimic.errors import MysqlError
from mysql_mimic.stream import MysqlStream


class MockReader:
    def __init__(self, data):
        self.data = data
        self.pos = 0

    async def read(self, size):
        start = self.pos
        end = self.pos + size
        self.pos = end

        return self.data[start:end]


class MockWriter:
    def __init__(self):
        self.data = b""

    def write(self, data):
        self.data += data

    async def drain(self):
        return


class TestMysqlStream(unittest.IsolatedAsyncioTestCase):
    def test_seq(self):
        s1 = MysqlStream(reader=None, writer=None)
        self.assertEqual(next(s1.seq), 0)
        self.assertEqual(next(s1.seq), 1)
        self.assertEqual(next(s1.seq), 2)
        s2 = MysqlStream(reader=None, writer=None)
        self.assertEqual(next(s2.seq), 0)
        self.assertEqual(next(s2.seq), 1)
        self.assertEqual(next(s2.seq), 2)

    def test_reset_seq(self):
        s = MysqlStream(reader=None, writer=None)
        self.assertEqual(next(s.seq), 0)
        self.assertEqual(next(s.seq), 1)
        s.reset_seq()
        self.assertEqual(next(s.seq), 0)
        self.assertEqual(next(s.seq), 1)

    async def test_bad_seq_read(self):
        reader = MockReader(b"\x00\x00\x00\x01")
        s = MysqlStream(reader=reader, writer=None)
        with self.assertRaises(MysqlError):
            await s.read()

    async def test_empty_read(self):
        reader = MockReader(b"\x00\x00\x00\x00")
        s = MysqlStream(reader=reader, writer=None)
        self.assertEqual(await s.read(), b"")
        self.assertEqual(next(s.seq), 1)

    async def test_small_read(self):
        reader = MockReader(b"\x01\x00\x00\x00k")
        s = MysqlStream(reader=reader, writer=None)
        self.assertEqual(await s.read(), b"k")
        self.assertEqual(next(s.seq), 1)

    async def test_medium_read(self):
        reader = MockReader(b"\xff\xff\x00\x00" + bytes(0xFFFF))
        s = MysqlStream(reader=reader, writer=None)
        self.assertEqual(await s.read(), bytes(0xFFFF))
        self.assertEqual(next(s.seq), 1)

    async def test_edge_read(self):
        reader = MockReader(b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x00\x00\x00\x01")
        s = MysqlStream(reader=reader, writer=None)
        self.assertEqual(await s.read(), bytes(0xFFFFFF))
        self.assertEqual(next(s.seq), 2)

    async def test_large_read(self):
        reader = MockReader(
            b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x06\x00\x00\x01" + b"kelsin"
        )
        s = MysqlStream(reader=reader, writer=None)
        self.assertEqual(await s.read(), bytes(0xFFFFFF) + b"kelsin")
        self.assertEqual(next(s.seq), 2)

    async def test_empty_write(self):
        writer = MockWriter()
        s = MysqlStream(reader=None, writer=writer)
        await s.write(b"")
        self.assertEqual(writer.data, b"\x00\x00\x00\x00")
        self.assertEqual(next(s.seq), 1)

    async def test_small_write(self):
        writer = MockWriter()
        s = MysqlStream(reader=None, writer=writer)
        await s.write(b"kelsin")
        self.assertEqual(writer.data, b"\x06\x00\x00\x00kelsin")
        self.assertEqual(next(s.seq), 1)

    async def test_medium_write(self):
        writer = MockWriter()
        s = MysqlStream(reader=None, writer=writer)
        await s.write(bytes(0xFFFF))
        self.assertEqual(writer.data, b"\xff\xff\x00\x00" + bytes(0xFFFF))
        self.assertEqual(next(s.seq), 1)

    async def test_edge_write(self):
        writer = MockWriter()
        s = MysqlStream(reader=None, writer=writer)
        await s.write(bytes(0xFFFFFF))
        self.assertEqual(
            writer.data, b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x00\x00\x00\x01"
        )
        self.assertEqual(next(s.seq), 2)

    async def test_large_write(self):
        writer = MockWriter()
        s = MysqlStream(reader=None, writer=writer)
        await s.write(bytes(0xFFFFFF) + b"kelsin")
        self.assertEqual(
            writer.data,
            b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x06\x00\x00\x01kelsin",
        )
        self.assertEqual(next(s.seq), 2)



import unittest
from unittest.mock import Mock

from mysql_mimic import Session
from mysql_mimic.charset import CharacterSet
from mysql_mimic.results import ResultSet
from mysql_mimic.admin import Admin


class TestAdmin(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        # pylint: disable=attribute-defined-outside-init
        self.session = Mock(spec=Session)
        self.session.show_columns.return_value = [
            {"name": "col1", "type": "INTEGER"},
            {"name": "col2", "type": "INTEGER"},
        ]
        self.cmd = Admin(connection_id=1, session=self.session)
        self.cmd.database = "db"

    async def test_parse_show_columns(self):
        for cmd, expected, result_length, col_length in [
            ("show columns from table", ("db", "table"), 2, 6),
            (
                " SHOW  EXTENDED  FULL  FIELDS  IN  `table`  IN  `db2` ",
                ("db2", "table"),
                2,
                9,
            ),
            ("show columns from table like '%2'", ("db", "table"), 1, 6),
            ("show columns from table like 'col1'", ("db", "table"), 1, 6),
            ("show columns from table like 'col%'", ("db", "table"), 2, 6),
            ("show columns from table like '%col%'", ("db", "table"), 2, 6),
            ("show columns from table like '_ol2'", ("db", "table"), 1, 6),
        ]:
            result = await self.cmd.parse(cmd)
            self.session.show_columns.assert_called_with(*expected)
            self.assertIsInstance(result, ResultSet)
            self.assertEqual(len(list(result.rows)), result_length)
            self.assertEqual(len(result.columns), col_length)

    async def test_parse_show_index(self):
        for cmd, result_length, col_length in [
            ("show index from table", 0, 15),
        ]:
            result = await self.cmd.parse(cmd)
            self.assertIsInstance(result, ResultSet)
            self.assertEqual(len(list(result.rows)), result_length)
            self.assertEqual(len(result.columns), col_length)

    async def test_parse_show_variables(self):
        for cmd, expected in [
            (
                "show variables",
                {
                    "version_comment": "mysql-mimic",
                    "version": "8.0.29",
                    "character_set_client": "utf8mb4",
                    "character_set_results": "utf8mb4",
                    "character_set_server": "utf8mb4",
                    "collation_server": "utf8mb4_general_ci",
                    "collation_database": "utf8mb4_general_ci",
                },
            ),
            (
                "SHOW  SESSION  VARIABLES  LIKE 'version_comment'",
                {"version_comment": "mysql-mimic"},
            ),
            (
                "SHOW  SESSION  VARIABLES  LIKE 'version_%'",
                {"version_comment": "mysql-mimic"},
            ),
            ("SET @@version_comment = 'hello'", {}),
            (
                "SHOW  SESSION  VARIABLES  LIKE 'version_comment'",
                {"version_comment": "hello"},
            ),
        ]:
            result = await self.cmd.parse(cmd)
            self.assertIsInstance(result, ResultSet)
            rows = {r[0]: r[1] for r in result.rows}
            self.assertEqual(rows, expected)

    async def test_parse_set_names(self):
        for cmd, expected_charset, expected_collation in [
            (
                "SET NAMES utf8",
                "utf8",
                "utf8_general_ci",
            ),
            (
                "SET NAMES big5 DEFAULT",
                "big5",
                "big5_chinese_ci",
            ),
            (
                "SET NAMES utf8 COLLATE utf8mb4_bin",
                "utf8",
                "utf8mb4_bin",
            ),
            (
                "set  names  'big5'  collate  'big5_chinese_ci'",
                "big5",
                "big5_chinese_ci",
            ),
        ]:
            await self.cmd.parse(cmd)
            self.assertEqual(
                self.cmd.variables["character_set_client"], expected_charset
            )
            self.assertEqual(
                self.cmd.variables["character_set_connection"], expected_charset
            )
            self.assertEqual(
                self.cmd.variables["character_set_results"], expected_charset
            )
            self.assertEqual(
                self.cmd.variables["collation_connection"], expected_collation
            )
            self.assertEqual(
                self.cmd.client_character_set, CharacterSet[expected_charset]
            )
            self.assertEqual(
                self.cmd.server_character_set, CharacterSet[expected_charset]
            )

    async def test_parse_set_variables(self):
        for cmd, expected in [
            ("SET version_comment = 'b'", "b"),
            ("SET @@version_comment = ON", True),
            ("SET  @@SESSION.version_comment  = OFF", False),
            ("set version_comment = DEFAULT", "mysql-mimic"),
            ("set version_comment = 1", 1),
            ("set version_comment = 2.3", 2.3),
            ("set version_comment = NULL", "mysql-mimic"),
        ]:
            await self.cmd.parse(cmd)
            self.assertEqual(self.cmd.variables["version_comment"], expected)



import unittest

from mysql_mimic.errors import MysqlError
from mysql_mimic.results import ensure_result_set


class TestResult(unittest.TestCase):
    def test_ensure_result_set__invalid(self):
        for result in [
            [1, 2],
            ([[1, 2]], ["a", "b"], ["a", "b"]),
            ([[1, 2]], [1, 2]),
        ]:
            with self.assertRaises(MysqlError):
                ensure_result_set(result)



import io
import contextlib
import unittest

from mysql_mimic import version


class TestTypes(unittest.TestCase):
    def test_version(self):
        self.assertTrue(isinstance(version.__version__, str))

    def test_main(self):
        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            version.main("__main__")

        self.assertEqual(f"{version.__version__}\n", out.getvalue())
